﻿using System;
using System.Drawing;
using System.Windows.Forms;
using SlimDX;

using DX11 = SlimDX.Direct3D11;
using DXGI = SlimDX.DXGI;

namespace TutorialSeries.DirectX10.Chapter1
{
    public partial class MainWindow : Form
    {
        private DX11.Device m_device;
        private DX11.DeviceContext m_deviceContext;
        private DX11.RenderTargetView m_renderTarget;
        private DXGI.SwapChain m_swapChain;
        private DXGI.SwapChainDescription m_swapChainDesc;
        private DXGI.Factory m_factory;
        private bool m_initialized;

        public MainWindow()
        {
            InitializeComponent();
        }

        /// <summary>
        /// Initializes device and other resources needed for rendering. Returns true, if successful.
        /// </summary>
        private bool Initialize3D()
        {
            try
            {
                m_device = new DX11.Device(
                    DX11.DriverType.Hardware,
                    DX11.DeviceCreationFlags.SingleThreaded);
                m_deviceContext = m_device.ImmediateContext;

                //Query for the dxgi adapter
                DXGI.Device dxgiDevice = new DXGI.Device(m_device);
                DXGI.Adapter dxgiAdapter = dxgiDevice.Adapter;
                m_factory = dxgiAdapter.GetParent<DXGI.Factory>();

                //Create the swapchain
                m_swapChainDesc = new DXGI.SwapChainDescription();
                m_swapChainDesc.OutputHandle = this.Handle;
                m_swapChainDesc.IsWindowed = true;
                m_swapChainDesc.BufferCount = 1;
                m_swapChainDesc.Flags = DXGI.SwapChainFlags.AllowModeSwitch;
                m_swapChainDesc.ModeDescription = new DXGI.ModeDescription(
                    this.Width,
                    this.Height,
                    new Rational(60, 1),
                    DXGI.Format.R8G8B8A8_UNorm);
                m_swapChainDesc.SampleDescription = new DXGI.SampleDescription(1, 0);
                m_swapChainDesc.SwapEffect = DXGI.SwapEffect.Discard;
                m_swapChainDesc.Usage = DXGI.Usage.RenderTargetOutput;

                m_swapChain = new DXGI.SwapChain(m_factory, m_device, m_swapChainDesc);

                DX11.Viewport viewPort = new DX11.Viewport();
                viewPort.X = 0;
                viewPort.Y = 0;
                viewPort.Width = this.Width;
                viewPort.Height = this.Height;
                viewPort.MinZ = 0f;
                viewPort.MaxZ = 1f;

                DX11.Texture2D backBuffer =
                    DX11.Texture2D.FromSwapChain<DX11.Texture2D>(m_swapChain, 0);
                m_renderTarget = new DX11.RenderTargetView(m_device, backBuffer);

                m_deviceContext.Rasterizer.SetViewports(viewPort);
                m_deviceContext.OutputMerger.SetTargets(m_renderTarget);

                m_initialized = true;
            }
            catch (Exception ex)
            {
                MessageBox.Show("Error while initializing Direct3D: \n" + ex.Message);
                m_initialized = false;
            }

            return m_initialized;
        }

        /// <summary>
        /// Rendering is done during the standard OnPaint event
        /// </summary>
        protected override void OnPaint(PaintEventArgs e)
        {
            base.OnPaint(e);

            if (m_initialized)
            {
                m_deviceContext.ClearRenderTargetView(m_renderTarget, new Color4(Color.CornflowerBlue));

                m_swapChain.Present(0, DXGI.PresentFlags.None);
            }
        }

        /// <summary>
        /// Initialize 3D-Graphics within OnLoad event
        /// </summary>
        protected override void OnLoad(EventArgs e)
        {
            base.OnLoad(e);
            Initialize3D();
        }
    }
}
