﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using SlimDX;

using DX11 = SlimDX.Direct3D11;
using DXGI = SlimDX.DXGI;

namespace TutorialSeries.DirectX10.Chapter5
{
    public partial class MainWindow : Form
    {
        private DX11.Device m_device;
        private DX11.DeviceContext m_deviceContext;
        private DXGI.SwapChainDescription m_swapChainDesc;
        private DXGI.SwapChain m_swapChain;
        private DXGI.Factory m_factory;
        private DX11.RenderTargetView m_renderTarget;
        private DX11.DepthStencilView m_renderTargetDepth;
        private bool m_initialized;

        private SimpleBox m_simpleBox;

        private Matrix m_viewMatrix;
        private Matrix m_projMatrix;
        private Matrix m_worldMatrix;
        private Matrix m_worldMatrix2;
        private Matrix m_viewProjMatrix;

        public MainWindow()
        {
            InitializeComponent();

            this.SetStyle(ControlStyles.ResizeRedraw, true);
            this.SetStyle(ControlStyles.AllPaintingInWmPaint, true);
            this.SetStyle(ControlStyles.Opaque, true);
        }

        /// <summary>
        /// Initialisiert unser 3D-Device und gibt true zurück, falls alles erfolgreich war
        /// </summary>
        private bool Initialize3D()
        {
            try
            {
                m_device = new DX11.Device(
                    DX11.DriverType.Hardware,
                    DX11.DeviceCreationFlags.SingleThreaded);
                m_deviceContext = m_device.ImmediateContext;

                //Query for the dxgi adapter
                DXGI.Device dxgiDevice = new DXGI.Device(m_device);
                DXGI.Adapter dxgiAdapter = dxgiDevice.Adapter;
                m_factory = dxgiAdapter.GetParent<DXGI.Factory>();

                m_swapChainDesc = new DXGI.SwapChainDescription();
                m_swapChainDesc.OutputHandle = this.Handle;
                m_swapChainDesc.IsWindowed = true;
                m_swapChainDesc.BufferCount = 1;
                m_swapChainDesc.Flags = DXGI.SwapChainFlags.AllowModeSwitch;
                m_swapChainDesc.ModeDescription = new DXGI.ModeDescription(
                    this.Width,
                    this.Height,
                    new Rational(60, 1),
                    DXGI.Format.R8G8B8A8_UNorm);
                m_swapChainDesc.SampleDescription = new DXGI.SampleDescription(1, 0);
                m_swapChainDesc.SwapEffect = DXGI.SwapEffect.Discard;
                m_swapChainDesc.Usage = DXGI.Usage.RenderTargetOutput;

                m_swapChain = new DXGI.SwapChain(m_factory, m_device, m_swapChainDesc);

                DX11.Viewport viewPort = new DX11.Viewport();
                viewPort.X = 0;
                viewPort.Y = 0;
                viewPort.Width = this.Width;
                viewPort.Height = this.Height;
                viewPort.MinZ = 0f;
                viewPort.MaxZ = 1f;

                DX11.Texture2D backBuffer =
                    DX11.Texture2D.FromSwapChain<DX11.Texture2D>(m_swapChain, 0);
                m_renderTarget = new DX11.RenderTargetView(m_device, backBuffer);

                m_deviceContext.Rasterizer.SetViewports(viewPort);
                m_deviceContext.OutputMerger.SetTargets(m_renderTarget);

                DX11.Texture2DDescription depthBufferDesc = new DX11.Texture2DDescription();
                depthBufferDesc.Width = this.Width;
                depthBufferDesc.Height = this.Height;
                depthBufferDesc.MipLevels = 1;
                depthBufferDesc.ArraySize = 1;
                depthBufferDesc.Format = DXGI.Format.D32_Float;
                depthBufferDesc.Usage = DX11.ResourceUsage.Default;
                depthBufferDesc.SampleDescription = new DXGI.SampleDescription(1, 0);
                depthBufferDesc.BindFlags = DX11.BindFlags.DepthStencil;
                depthBufferDesc.CpuAccessFlags = DX11.CpuAccessFlags.None;
                depthBufferDesc.OptionFlags = DX11.ResourceOptionFlags.None;

                DX11.Texture2D depthBuffer = new DX11.Texture2D(m_device, depthBufferDesc);
                m_renderTargetDepth = new DX11.DepthStencilView(m_device, depthBuffer);

                m_deviceContext.Rasterizer.SetViewports(viewPort);
                m_deviceContext.OutputMerger.SetTargets(m_renderTargetDepth, m_renderTarget);

                m_viewMatrix = Matrix.LookAtLH(
                     new Vector3(0f, 0f, -4f),
                     new Vector3(0f, 0f, 1f),
                     new Vector3(0f, 1f, 0f));
                m_projMatrix = Matrix.PerspectiveFovLH(
                    (float)Math.PI * 0.5f,
                    this.Width / (float)this.Height,
                    0.1f, 100f);
                m_viewProjMatrix = m_viewMatrix * m_projMatrix;
                m_worldMatrix = Matrix.RotationYawPitchRoll(0.85f, 0.85f, 0f);
                m_worldMatrix2 = Matrix.Translation(0f, 3f, 0f) * Matrix.RotationYawPitchRoll(0.85f, 0.85f, 0f);   

                m_simpleBox = new SimpleBox();
                m_simpleBox.LoadResources(m_device);

                m_initialized = true;
            }
            catch (Exception ex)
            {
                MessageBox.Show("Error while initializing Direct3D10: \n" + ex.Message);
                m_initialized = false;
            }

            return m_initialized;
        }

        /// <summary>
        /// Rendering is done during the standard OnPaint event
        /// </summary>
        protected override void OnPaint(PaintEventArgs e)
        {
            base.OnPaint(e);

            if (m_initialized)
            {
                m_deviceContext.ClearRenderTargetView(m_renderTarget, new Color4(Color.CornflowerBlue));
                m_deviceContext.ClearDepthStencilView(m_renderTargetDepth, DX11.DepthStencilClearFlags.Depth, 1f, 0);

                m_simpleBox.Render(m_device, m_worldMatrix, m_viewProjMatrix);
                m_simpleBox.Render(m_device, m_worldMatrix2, m_viewProjMatrix);

                m_swapChain.Present(0, DXGI.PresentFlags.None);
            }
        }

        /// <summary>
        /// Initialize 3D-Graphics within OnLoad event
        /// </summary>
        protected override void OnLoad(EventArgs e)
        {
            base.OnLoad(e);
            Initialize3D();
        }
    }
}
