﻿using System;
using System.Runtime.InteropServices;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using SlimDX;

using DX11 = SlimDX.Direct3D11;
using DXGI = SlimDX.DXGI;
using D3DCompiler = SlimDX.D3DCompiler;

namespace TutorialSeries.DirectX10.Chapter6
{
    public class Stars
    {
        private const int STARS_COUNT = 50;

        private List<StarInstance> m_stars;
        private Random m_randomizer;
        private DX11.Buffer m_vertexBuffer;
        private DX11.InputLayout m_vertexLayout;
        private DX11.Effect m_effect;
        private DX11.EffectTechnique m_effectTechnique;
        private DX11.EffectPass m_effectPass;
        private DX11.EffectMatrixVariable m_transformVariable;
        private DX11.EffectVectorVariable m_colorVariable;
        private DX11.Texture2D m_texture;
        private DX11.EffectResourceVariable m_resourceVariable;
        private DX11.ShaderResourceView m_textureView;
        private DX11.BlendState m_blendState;

        public void LoadResources(DX11.Device device)
        {
            Vertex[] vertices = new Vertex[]
            {
                new Vertex(new Vector3(-1f, -1f, 0f), new Vector2(0, 0)),
                new Vertex(new Vector3(-1f, 1f, 0f), new Vector2(0, 1f)),
                new Vertex(new Vector3(1f, 1f, 0f), new Vector2(1f, 1f)),
                new Vertex(new Vector3(1f, 1f, 0f), new Vector2(1f, 1f)),
                new Vertex(new Vector3(1f, -1f, 0f), new Vector2(1f, 0)),
                new Vertex(new Vector3(-1f, -1f, 0f), new Vector2(0, 0))
            };

            DataStream outStream = new DataStream(6 * Marshal.SizeOf(typeof(Vertex)), true, true);
            outStream.WriteRange(vertices);
            outStream.Position = 0;

            DX11.BufferDescription bufferDescription = new DX11.BufferDescription();
            bufferDescription.BindFlags = DX11.BindFlags.VertexBuffer;
            bufferDescription.CpuAccessFlags = DX11.CpuAccessFlags.None;
            bufferDescription.OptionFlags = DX11.ResourceOptionFlags.None;
            bufferDescription.SizeInBytes = 6 * Marshal.SizeOf(typeof(Vertex));
            bufferDescription.Usage = DX11.ResourceUsage.Default;

            m_vertexBuffer = new DX11.Buffer(device, outStream, bufferDescription);
            outStream.Close();

            D3DCompiler.ShaderBytecode shaderByteCode =
               D3DCompiler.ShaderBytecode.CompileFromFile(
                   "Shaders/SimpleRendering.fx", "fx_5_0",
                   D3DCompiler.ShaderFlags.None, D3DCompiler.EffectFlags.None);
            m_effect = new DX11.Effect(
                device, shaderByteCode);
            m_effectTechnique = m_effect.GetTechniqueByIndex(0);
            m_effectPass = m_effectTechnique.GetPassByIndex(0);
            m_transformVariable = m_effect.GetVariableByName("WorldViewProj").AsMatrix();
            m_resourceVariable = m_effect.GetVariableByName("Texture").AsResource();
            m_colorVariable = m_effect.GetVariableByName("StarColor").AsVector();

            DX11.InputElement[] inputElements = new DX11.InputElement[]
            {
                new DX11.InputElement("POSITION",0,DXGI.Format.R32G32B32_Float,0,0),
                new DX11.InputElement("TEXCOORD",0,DXGI.Format.R32G32_Float,12,0)
            };
            m_vertexLayout = new DX11.InputLayout(device, inputElements, m_effectPass.Description.Signature);

            m_texture = DX11.Texture2D.FromFile(device, "Textures/Star.bmp");
            m_textureView = new DX11.ShaderResourceView(device, m_texture);

            m_stars = new List<StarInstance>();
            m_randomizer =new Random(Environment.TickCount);
            for (int loop = 1; loop <= STARS_COUNT; loop++)
            {
                StarInstance newInstance = new StarInstance();
                newInstance.Angle = 0f;
                newInstance.Distance = ((float)loop / (float)STARS_COUNT) * 5f;
                newInstance.Color = new Vector4(
                        m_randomizer.Next(128, 255) / 255f,
                        m_randomizer.Next(128, 255) / 255f,
                        m_randomizer.Next(128, 255) / 255f,
                    1f);
                m_stars.Add(newInstance);
            }

            DX11.BlendStateDescription blendDesc = new DX11.BlendStateDescription();
            blendDesc.RenderTargets[0].BlendOperation = DX11.BlendOperation.Add;
            blendDesc.RenderTargets[0].BlendOperationAlpha = DX11.BlendOperation.Add;
            blendDesc.RenderTargets[0].SourceBlendAlpha = DX11.BlendOption.Zero;
            blendDesc.RenderTargets[0].DestinationBlendAlpha = DX11.BlendOption.Zero;
            blendDesc.RenderTargets[0].SourceBlend = DX11.BlendOption.SourceColor;
            blendDesc.RenderTargets[0].DestinationBlend = DX11.BlendOption.One;
            blendDesc.AlphaToCoverageEnable = false;
            blendDesc.RenderTargets[0].BlendEnable = true;
            blendDesc.RenderTargets[0].RenderTargetWriteMask = DX11.ColorWriteMaskFlags.All;
            m_blendState = DX11.BlendState.FromDescription(device, blendDesc);
        }

        public void Render(DX11.Device device, Matrix world, Matrix viewProj)
        {
            DX11.DeviceContext deviceContext = device.ImmediateContext;

            deviceContext.InputAssembler.InputLayout = m_vertexLayout;
            deviceContext.InputAssembler.PrimitiveTopology = DX11.PrimitiveTopology.TriangleList;
            deviceContext.InputAssembler.SetVertexBuffers(0, new DX11.VertexBufferBinding(m_vertexBuffer, Marshal.SizeOf(typeof(Vertex)), 0));

            m_resourceVariable.SetResource(m_textureView);
            deviceContext.OutputMerger.BlendState = m_blendState;

            for(int loop=0 ; loop<STARS_COUNT; loop++)
            {
                StarInstance actStar = m_stars[loop];

                m_colorVariable.Set(actStar.Color);
                m_transformVariable.SetMatrix((world * Matrix.Translation(actStar.Distance, 0f, 0f) * Matrix.RotationZ(actStar.Angle)) * viewProj);

                m_effectPass.Apply(deviceContext);
                deviceContext.Draw(6, 0);

                actStar.Angle += ((float)loop / (float)STARS_COUNT * 2f % 1f) / 57f;
                if (actStar.Angle > (float)Math.PI * 2f) { actStar.Angle = actStar.Angle - (float)Math.PI * 2f; }

                actStar.Distance = actStar.Distance - 0.01f;
                if (actStar.Distance < 0f)
                {
                    actStar.Distance = 5.0f;
                    actStar.Color = new Vector4(
                        m_randomizer.Next(128, 255) / 255f,
                        m_randomizer.Next(128, 255) / 255f,
                        m_randomizer.Next(128, 255) / 255f,
                        1f);
                }
            }
        }

        //*********************************************************************
        //*********************************************************************
        //*********************************************************************
        class StarInstance
        {
            public Vector4 Color;
            public float Angle;
            public float Distance;
        }
    }
}
